import numpy
import pandas
import operator
import math
import csv
import copy
import warnings

from primalytics.tools import linear, reciprocal, logit, numeric_types, \
    na_values


class Emblem:
    def __init__(self, filename: str = None, debug=False, sep: str = ',',
                 sheet_name=0):
        self.__is_model_present = False
        self.__debug = debug
        self.__file = []
        self.__file_len = None

        # Coefficients ---------------------------------------------------------
        self.base_weight = None

        # Rows -----------------------------------------------------------------
        self.__start_univariate_row = 0
        self.__last_univariate_row = 0

        self.univariate_beta = {}
        self.bivariate_beta = {}

        self.__closing_str = 'Orthogonal Polynomial Equations'

        # Variables ------------------------------------------------------------
        self.__row = 0
        self.__end_bivariates_reached = False

        self.__operations = {
            '+': operator.add,
            '-': operator.sub,
            '*': operator.mul,
            '/': operator.truediv
        }
        self.__link_functions = {
            'exp': numpy.exp,
            'log': numpy.log,
            'linear': numpy.vectorize(linear),
            'reciprocal': numpy.vectorize(reciprocal),
            'logit': logit
        }

        # Startup functions ----------------------------------------------------
        if filename is not None:
            self.__read_input_file(filename, sep, sheet_name)
            self.__read_base_level()
            self.__read_univariates()
            self.__read_bivariates()
            self.__flatten_bivariates()
            self.__is_model_present = True

    def __read_input_file(self, filename, sep, sheet_name):
        ext = filename.split('.')[-1]
        if ext == 'csv':
            with open(filename, 'rt') as csvfile:
                spamreader = csv.reader(csvfile, quotechar='"', delimiter=sep)
                for row in spamreader:
                    self.__file.append(numpy.array(row).astype(str))
        elif ext in ('xlsx', 'xls'):
            self.__file = list(map(numpy.array, pandas.read_excel(
                io=filename, sheet_name=sheet_name, keep_default_na=False,
                na_values=na_values, header=0).fillna('').values))
        else:
            raise Exception('Invalid extension \'{}\' '
                            'for Emblem filename'.format(ext))
        self.__file_len = len(self.__file)

    def __read_base_level(self):
        self.base_weight = float(self.__file[0][2])
        self.__start_univariate_row = 5

    def __read_univariates(self):
        if self.__file[5][0] != '':
            valid_indices = numpy.where(
                self.__file[self.__start_univariate_row] != '')[0]
            univariate_levels_cols = numpy.array(valid_indices)
            self.univariate_beta = {}
            self.__last_univariate_row = 0

            for col in univariate_levels_cols:
                single_univariate_beta = {}
                self.__row = self.__start_univariate_row + 1
                single_univariate_end_reached = False
                univariate_name = self.__file[self.__row - 1][col]
                if self.__debug:
                    print('{:-<100}'.format('{} '.format(univariate_name)))

                while not single_univariate_end_reached:

                    if self.__file_len == self.__row + 1:
                        single_univariate_end_reached = True
                        self.__end_bivariates_reached = True
                    elif self.__file[self.__row + 1][col + 1] == '':
                        single_univariate_end_reached = True
                    level = pandas.to_numeric(
                        arg=self.__file[self.__row][col],
                        errors='ignore',
                        downcast='integer')

                    if self.__debug:
                        print('\trow: {:<6}| col: {:<6}| level: {:<40}| value: '
                              '{:<8}'.format(
                                    self.__row, col, level,
                                    self.__file[self.__row][col + 1]))

                    single_univariate_beta.update(
                        {level: float(self.__file[self.__row][col + 1])})
                    self.__row += 1

                self.univariate_beta.update(
                    {univariate_name: single_univariate_beta})
                self.__last_univariate_row = \
                    max(self.__last_univariate_row, self.__row)
        else:
            self.__last_univariate_row = self.__start_univariate_row

    def __read_bivariates(self):
        self.__row = self.__last_univariate_row + 2
        self.bivariate_beta = {}

        if not self.__end_bivariates_reached:
            if not (self.__file[self.__row - 2][0] == self.__closing_str
                    or self.__file[self.__row - 1][0] == self.__closing_str
                    or self.__file[self.__row][0] == self.__closing_str):
                self.__cycle_over_bivariates()

    def __cycle_over_bivariates(self):
        while not self.__end_bivariates_reached:
            first_bivariate_name = self.__file[self.__row][2]
            second_bivariate_name = self.__file[self.__row + 2][0]

            bivariate_interaction_name = \
                first_bivariate_name + '|' + second_bivariate_name

            valid_indices = numpy.where(self.__file[self.__row + 1] != '')
            second_bivariate_levels = pandas.to_numeric(
                arg=self.__file[self.__row + 1][valid_indices],
                errors='ignore',
                downcast='integer')

            self.__row += 2

            # cycle over bivariate matrix's rows
            single_bivariate_end_reached = False
            single_bivariate_beta = {}
            while not single_bivariate_end_reached:
                try:
                    second_bivariate_level = pandas.to_numeric(
                        arg=self.__file[self.__row][1],
                        errors='ignore',
                        downcast='integer')
                except ValueError:
                    second_bivariate_level = self.__file[self.__row][1]

                row_values = pandas.to_numeric(
                    self.__file[self.__row][2:2 + len(second_bivariate_levels)])

                single_bivariate_beta.update(
                    {second_bivariate_level: dict(
                        zip(second_bivariate_levels, row_values))})

                if self.__row >= self.__file_len - 1:
                    single_bivariate_end_reached = True
                    self.__end_bivariates_reached = True
                elif self.__file[self.__row + 1][2] == '':
                    single_bivariate_end_reached = True

                    if self.__row < self.__file_len - 1:
                        if self.__file[self.__row + 2][0] == self.__closing_str:
                            self.__end_bivariates_reached = True

                self.__row += 1

            self.__row += 3

            single_bivariate_beta = pandas.DataFrame(single_bivariate_beta)
            single_bivariate_beta = single_bivariate_beta.T.to_dict()

            self.bivariate_beta.update(
                {bivariate_interaction_name: single_bivariate_beta})

    def __flatten_bivariates(self):
        self.__bivariate_beta_flat = {}
        for interaction_name, bivariate_dict in self.bivariate_beta.items():
            bivariate_df = pandas.DataFrame(bivariate_dict)

            if bivariate_df.index.dtype in numeric_types:
                out_format = '{:0.3f}'
            else:
                out_format = '{}'

            bivariate_df.index = \
                [out_format.format(x) for x in bivariate_df.index]

            if bivariate_df.columns.dtype in numeric_types:
                out_format = '{:0.3f}'
            else:
                out_format = '{}'

            bivariate_df.columns = \
                [out_format.format(x) for x in bivariate_df.columns]

            self.__bivariate_beta_flat.update({
                interaction_name: {
                    '{}|{}'.format(k[0], k[1]): v
                    for k, v in bivariate_df.unstack().items()
                }
            })

    def __check_model_presence(self):
        if not self.__is_model_present:
            raise Exception('Error! No model present in class Emblem. use set_'
                            'univariates(...), set_bivariates(...) '
                            'methods to insert values')

    def __get_univariate_beta(self, df, variable, debug=False):
        if debug:
            try:
                result = df[variable].apply(
                    lambda x: self.univariate_beta[variable][x])
            except KeyError as ke:
                print(variable, ke)
                raise KeyError
        else:
            result = df[variable].map(self.univariate_beta[variable])

        return result

    def __get_bivariate_beta(self, df, interaction_name, debug=False,
                             return_levels: bool = False):

        first_bivariate = interaction_name.split('|')[0]
        second_bivariate = interaction_name.split('|')[1]

        first_bivariate_values = df[first_bivariate]
        second_bivariate_values = df[second_bivariate]

        if first_bivariate_values.dtype in numeric_types:
            first_bivariate_values = first_bivariate_values.apply(
                lambda x: '{:0.3f}'.format(x))
        else:
            first_bivariate_values = first_bivariate_values.astype(str)

        if second_bivariate_values.dtype in numeric_types:
            second_bivariate_values = second_bivariate_values.apply(
                lambda x: '{:0.3f}'.format(x))
        else:
            second_bivariate_values = second_bivariate_values.astype(str)

        compact_bivariate = \
            first_bivariate_values + \
            pandas.Series(['|' for _ in range(len(df))], index=df.index) + \
            second_bivariate_values

        if debug:
            try:
                values = compact_bivariate.apply(
                    lambda x: self.__bivariate_beta_flat[interaction_name][x])
            except KeyError as ke:
                try:
                    ke = float(str(ke))
                except (TypeError, ValueError):
                    pass
                else:
                    if math.isnan(ke):
                        error_condition = compact_bivariate.apply(
                            lambda x: math.isnan(x)
                            if isinstance(x, float)
                            else False)
                    else:
                        error_condition = compact_bivariate.isin([ke])

                    print(interaction_name, ke, '\n')
                    print(pandas.concat(
                        [first_bivariate_values, second_bivariate_values,
                         compact_bivariate], axis=1)[error_condition]
                          .iloc[0:10, :])
                raise KeyError
        else:
            values = compact_bivariate.map(
                self.__bivariate_beta_flat[interaction_name])

        if return_levels:
            return values, compact_bivariate
        else:
            return values

    def __check_base_weight(self, base_weight):
        if base_weight is None:
            if self.base_weight is None:
                raise Exception('Error! base_waight is not set. '
                                'Specify its value using base_weight=<value>')
        else:
            self.base_weight = base_weight

    def set_univariates(self, univariates: dict, base_weight=None):
        self.__check_base_weight(base_weight=base_weight)

        self.univariate_beta = univariates
        self.__flatten_bivariates()
        self.__is_model_present = True

    def set_bivariates(self, bivariates: dict, base_weight=None):
        self.__check_base_weight(base_weight=base_weight)

        self.bivariate_beta = bivariates
        self.__flatten_bivariates()
        self.__is_model_present = True

    def linear_predictor(self, df: pandas.DataFrame, vars_subset: list = None,
                         excluded_vars=None, return_levels: bool = False,
                         debug=False) -> numpy.array:
        if return_levels:
            coeffs_df, levels_df = self.get_coeffs(
                df=df,
                vars_subset=vars_subset,
                excluded_vars=excluded_vars,
                return_levels=return_levels,
                debug=debug
            )

            values_sum = coeffs_df.sum(axis=1, skipna=False).values
            return values_sum + self.base_weight, coeffs_df, levels_df
        else:
            return self.get_coeffs(
                df=df,
                vars_subset=vars_subset,
                excluded_vars=excluded_vars,
                return_levels=return_levels,
                debug=debug
            ).sum(axis=1, skipna=False).values + self.base_weight

    def product(self, df: pandas.DataFrame, vars_subset: list = None,
                excluded_vars=None, return_levels: bool = False,
                debug=False) -> numpy.array:

        if return_levels:
            coeffs_df, levels_df = self.get_coeffs(
                df=df,
                vars_subset=vars_subset,
                excluded_vars=excluded_vars,
                return_levels=return_levels,
                debug=debug
            )
            values_prod = coeffs_df.prod(axis=1, skipna=False).values
            return values_prod, coeffs_df, levels_df
        else:
            return self.get_coeffs(
                df=df,
                vars_subset=vars_subset,
                excluded_vars=excluded_vars,
                return_levels=return_levels,
                debug=debug).prod(axis=1, skipna=False).values

    def get_coeffs(self, df: pandas.DataFrame, vars_subset: list = None,
                   excluded_vars: list = None, return_levels: bool = False,
                   debug=False) -> numpy.array:

        self.__check_model_presence()

        coeffs_df = pandas.DataFrame(index=df.index)
        levels_df = pandas.DataFrame(index=df.index)

        if excluded_vars is None:
            excluded_vars = []

        if vars_subset is not None:
            excluded_vars = \
                [x for x in self.univariate_beta.keys() if x not in vars_subset]
            bivs = [x.split('|') for x in self.bivariate_beta.keys()]

            for biv in bivs:
                if biv[0] not in vars_subset and biv[1] not in vars_subset:
                    excluded_vars += biv
                elif biv[0] not in vars_subset and biv[1] in vars_subset:
                    raise Exception('Also variable \'{0}\' must appear in var_'
                                    'subset as \'{1}\' appears in it and '
                                    'bivariate \'{0}|{1}\' exists'
                                    .format(biv[0], biv[1]))
                elif biv[0] in vars_subset and biv[1] not in vars_subset:
                    raise Exception('Also variable \'{1}\' must appear in var_'
                                    'subset as \'{0}\' appears in it and '
                                    'bivariate \'{0}|{1}\' exists'
                                    .format(biv[0], biv[1]))

        for variable in self.univariate_beta.keys():
            if variable not in excluded_vars:
                coeffs = self.__get_univariate_beta(
                    df=df,
                    variable=variable,
                    debug=debug).values

                coeffs_df[variable] = coeffs
                if return_levels:
                    levels_df[variable] = df[variable]

        for interaction_name in self.__bivariate_beta_flat.keys():
            first_bivariate = interaction_name.split('|')[0]
            second_bivariate = interaction_name.split('|')[1]

            if not ((first_bivariate in excluded_vars)
                    or (second_bivariate in excluded_vars)):
                if return_levels:
                    coeffs, compact_bivariate = self.__get_bivariate_beta(
                        df=df,
                        interaction_name=interaction_name,
                        debug=debug,
                        return_levels=return_levels
                    )

                    levels_df[interaction_name] = compact_bivariate
                else:
                    coeffs = self.__get_bivariate_beta(
                        df=df,
                        interaction_name=interaction_name,
                        debug=debug,
                        return_levels=return_levels
                    )

                coeffs_df[interaction_name] = coeffs.values

        null_values = coeffs_df.isna().sum().sum()

        if null_values > 0:
            warnings.warn(
                '{} null values found in result'.format(null_values), Warning)

        if return_levels:
            return coeffs_df, levels_df
        else:
            return coeffs_df

    def predict(self, df: pandas.DataFrame, link_function, excluded_vars=None,
                debug=False) -> pandas.Series:
        self.__check_model_presence()

        if isinstance(link_function, str):
            link_function = self.__link_functions[link_function]
        if link_function is None:
            exc = 'Error! Invalid link_function specified! '
            'Valid values are \'{}\''.format(
                "', '".join(list(self.__link_functions.keys())))

            raise Exception(exc)
        else:
            link_function = numpy.vectorize(link_function)

        linear_predictor = self.linear_predictor(
            df=df,
            excluded_vars=excluded_vars,
            debug=debug
        )

        result = pandas.Series(
            data=link_function(linear_predictor),
            index=df.index)

        return result

    def get_beta(self, df: pandas.DataFrame, vars_subset: list = None):
        if vars_subset is None:
            vars_subset = list(df.columns)

        self.__check_model_presence()

        beta_values = pandas.DataFrame(index=df.index)
        beta_values['base_weight'] = self.base_weight

        levels = pandas.DataFrame(index=df.index)

        for variable, beta_dict in self.univariate_beta.items():
            if variable in vars_subset:
                values = self.__get_univariate_beta(
                    df=df,
                    variable=variable
                )

                beta_values[variable] = values
                levels[variable] = df[variable]

        for interaction_name in self.bivariate_beta.keys():
            first_bivariate = interaction_name.split('|')[0]
            second_bivariate = interaction_name.split('|')[1]

            if first_bivariate in vars_subset \
                    and second_bivariate in vars_subset:
                values, compact_bivariate = self.__get_bivariate_beta(
                    df=df,
                    interaction_name=interaction_name,
                    debug=False,
                    return_levels=True
                )

                beta_values[interaction_name] = values
                levels[interaction_name] = compact_bivariate

        return levels, beta_values

    def rescale(self, variables_with_levels: dict, rescale_to: int):
        """
        :param variables_with_levels: dict with variables to rescale and levels
            to set as base levels
        :param rescale_to: value to set for the base level: valid values {0, 1}
        """
        self.__check_model_presence()

        if rescale_to == 1:
            operation = operator.truediv
            inverse_operation = operator.mul
        elif rescale_to == 0:
            operation = operator.sub
            inverse_operation = operator.add
        else:
            raise Exception('Invalid recaling value \'{}\'. '
                            'Valid values are {{0, 1}}'.format(rescale_to))

        rescaled = self.copy()

        for variable, base_level in variables_with_levels.items():
            if variable in self.univariate_beta:
                try:
                    rescaling_value = self.univariate_beta[variable][base_level]
                except KeyError as ke:
                    raise Exception('KeyError! invalid key \'{}\' for variable '
                                    '\'{}\''.format(ke, variable))

                rescaled.univariate_beta[variable] = {
                    k: operation(v, rescaling_value)
                    for k, v in rescaled.univariate_beta[variable].items()}

                rescaled.base_weight = \
                    inverse_operation(rescaled.base_weight, rescaling_value)

            elif variable in self.bivariate_beta:
                try:
                    bivariate_dict = self.bivariate_beta[variable]
                    rescaling_value = \
                        bivariate_dict[base_level[0]][base_level[1]]
                except KeyError as ke:
                    raise Exception('KeyError! invalid key {} for variable '
                                    '{}'.format(ke, variable))
                rescaled.bivariate_beta[variable] = {
                    k1: {
                        k2: operation(v2, rescaling_value)
                        for k2, v2 in v1.items()
                    } for k1, v1 in rescaled.bivariate_beta[variable].items()
                }
            else:
                raise Exception('KeyError! variable \'{}\' not found in '
                                'univariate or bivariate dict'.format(variable))

        return rescaled

    def copy(self, deep: bool = True):
        if deep:
            return copy.deepcopy(self)
        else:
            return copy.copy(self)

    def write_to_file(self, filename: str):
        ext = filename.split('.')[-1]
        if ext == 'xlsx':
            shift = 1
        else:
            shift = 0

        self.__check_model_presence()

        base_weight_len = 3 + shift

        n_univariates = len(self.univariate_beta)
        if n_univariates > 0:
            lengths = [len(x) for x in self.univariate_beta.values()]
            n_max_univariate_levels = max(lengths)
            univariate_len = 3 + n_max_univariate_levels
        else:
            univariate_len = 0

        n_bivariates = len(self.bivariate_beta)
        if n_bivariates > 0:
            lengths = [len(x) for x in self.bivariate_beta.values()]
            bivariate_len = 5 * n_bivariates + sum(lengths) - 1
            n_max_bivariate_leves = max([
                len(x[list(x.keys())[0]])
                for x in self.bivariate_beta.values()
            ])
        else:
            n_max_bivariate_leves = 0
            bivariate_len = 0

        file_width = max(3 * n_univariates - 1, n_max_bivariate_leves + 1)
        file_length = base_weight_len + univariate_len + bivariate_len

        empty_list = ['' for _ in range(file_width)]
        file_array = [empty_list for _ in range(file_length)]

        file_array[shift][0] = 'Base'
        file_array[shift][2] = self.base_weight

        for (i, (variable, levels_values)) \
                in enumerate(self.univariate_beta.items()):

            file_array[shift + 3][3 * i] = variable
            file_array[shift + 5][3 * i] = variable

            for (j, (level, value)) in enumerate(levels_values.items()):
                file_array[shift + 6 + j][3 * i] = level
                file_array[shift + 6 + j][3 * i + 1] = value

        previous_bivariates_len = 0

        for (i, (interaction, matrix)) \
                in enumerate(self.bivariate_beta.items()):

            first_bivariate, second_bivariate = interaction.split('|')
            row = base_weight_len + univariate_len + previous_bivariates_len
            file_array[row][0] = '{} * {}'.format(
                first_bivariate, second_bivariate)
            file_array[row + 2][2] = first_bivariate
            file_array[row + 4][0] = second_bivariate

            n_second_bivariate_levels = len(matrix[list(matrix.keys())[0]])
            file_array[row + 3][2:2 + n_second_bivariate_levels] = \
                [k for k in matrix[list(matrix.keys())[0]].keys()]

            for (j, (level, levels_values)) in enumerate(matrix.items()):
                file_array[row + 4 + j][1] = level
                file_array[row + 4 + j][2:2 + n_second_bivariate_levels] = \
                    [v for v in levels_values.values()]

            previous_bivariates_len += 5 + len(self.bivariate_beta[interaction])

        with open(filename, 'w') as f:
            f.writelines([','.join(map(str, x)) + '\n' for x in file_array])



import pandas
import numpy
import re
from scipy import stats
import math
import warnings

from collections import Counter


def take_last(s: pandas.Series):
    return s.iloc[-1]


def take_first(s: pandas.Series):
    return s.iloc[0]


def count_unique(x: pandas.Series):
    return len(x.unique())


def bin_variable(s: pandas.Series, min_value, max_value, width, how, default=None, lower=None, upper=None, errors=None):
    """
    bin pandas Series
    :param s: pandas Series
    :param min_value: minimum bining value
    :param max_value: maximum binning value
    :param width: bin width
    :param how: function to use for binning. Use numpy.ceil, numpy.floor
    :param default: default value to replace nans
    :param lower: value to assign to values smaller than min_value
    :param upper: value to assign to values larger  than max_value
    :param errors: error values in the Series to be treated a if they where nans
    :return: binned series s_binned
    """
    s_binned = s.copy()
    nan_values = numpy.array(s_binned.isna()) | numpy.array((s_binned == errors))

    conditions = numpy.array([True for _ in range(len(s_binned))])

    if lower is not None:
        s_binned.loc[s_binned < min_value] = lower
        conditions = conditions & (s_binned != lower) & (min_value <= s_binned)

    if upper is not None:
        s_binned.loc[s_binned > max_value] = upper
        conditions = conditions & (s_binned != upper) & (s_binned <= max_value)

    s_binned.loc[conditions] = list(numpy.clip(how(s_binned.loc[conditions] / width) * width, min_value, max_value))
    s_binned = numpy.round(s_binned, 8)
    if default is not None:
        s_binned.loc[nan_values] = default

    return s_binned


def detect_outliers(df, n, features):
    """
    Takes a dataframe df of features and returns a list of the indices
    corresponding to the observations containing more than n outliers according
    to the Tukey method.
    """
    outlier_indices = []

    # iterate over features(columns)
    for col in features:
        # 1st quartile (25%)
        q1 = numpy.percentile(df[col].dropna(), 25)
        # 3rd quartile (75%)
        q3 = numpy.percentile(df[col].dropna(), 75)
        # Interquartile range (iqr)
        iqr = q3 - q1

        # outlier step
        outlier_step = 1.5 * iqr

        # Determine a list of indices of outliers for feature col
        outlier_list_col = df[(df[col] < q1 - outlier_step) | (df[col] > q3 + outlier_step)].index

        # append the found outlier indices for col to the list of outlier indices
        outlier_indices.extend(outlier_list_col)

    # select observations containing more than 2 outliers
    outlier_indices = Counter(outlier_indices)
    multiple_outliers = list(k for k, v in outlier_indices.items() if v > n)

    return multiple_outliers


def multiple_replace(s: pandas.Series, replacing_dict: dict, default=None):
    """
    replaces multiple values at once
    :param s: pandas Series to act on
    :param replacing_dict: dictionary with replacements. Has to be in the form: {'<replacing_value_1>' : [<list_of_values_to_replace], '<replacing_value_2>' : [<list_of_values_to_replace] ...}
    :param default: default value. used to fill values not in replacing_value_n
    :return:
    """
    for key, value in replacing_dict.items():
        s = s.replace(value, key)
    if default is not None:
        s = s.map(dict(zip(replacing_dict.keys(), replacing_dict.keys()))).fillna(default)
    return s


def find_regex(regex, value):
    if re.search(regex, value) is not None:
        return True
    else:
        return False


def fill_empty_with_distr(s: pandas.Series, values_to_fill: list, values_to_avoid: list = None):
    """
    :param s: pandas Series to be filled
    :param values_to_fill: value to fill
    :param values_to_avoid: values not valid for filling
    :return:
    """

    compl_values_to_avoid = [numpy.nan]
    if values_to_avoid is not None:
        compl_values_to_avoid += values_to_avoid

    filling_index = s.index[s.isin(values_to_fill)]

    distrib_data = s.value_counts()
    distrib_data = distrib_data[~distrib_data.index.isin(values_to_fill + compl_values_to_avoid)]
    distrib_data_dict = {k: v for k, v in enumerate(distrib_data.index)}

    xk, pk = list(distrib_data_dict.keys()), distrib_data.values / distrib_data.values.sum()
    distrib = stats.rv_discrete(values=(xk, pk))

    filled_s = s.copy()
    filled_s[filling_index] = [distrib_data_dict[distrib.rvs()] for _ in filling_index]

    return filled_s


def fill_str(s: pandas.Series, length, fill_value: str = '0', fillna=False):
    if s.isna().any() and not fillna:
        warnings.warn('{} NaNs found in input series. Mapping skipped'.format(sum(s.isna())), Warning)

    return s.astype(float).apply(lambda x: str(int(x)).rjust(length, fill_value) if not math.isnan(x) else ''.rjust(length, fill_value) if fillna else x)



import pandas
import numpy
import matplotlib.pyplot as plt
import seaborn as sns
import json

float_types = (float, numpy.float, numpy.float16, numpy.float32, numpy.float64)
int_types = (int, numpy.int, numpy.int16, numpy.int32, numpy.int64)
numeric_types = float_types + int_types

na_values = [
	'', '#N/A', '#N/A N/A', '#NA', '-1.#IND', '-1.#QNAN', '-NaN', '-nan',
	'1.#IND', '1.#QNAN', 'N/A', 'NULL', 'NaN', 'n/a', 'nan', 'null'
]


def get_json_dumpable(data: dict, depth=0, max_depth=10) -> dict:
	prc_data = {}
	for k, v in data.items():
		if isinstance(v, (str, bool, pandas.datetime,)):
			v_new = str(v)
		elif isinstance(v, float_types):
			v_new = float(v)
		elif isinstance(v, int_types):
			v_new = int(v)
		elif isinstance(v, dict):
			if depth == max_depth:
				raise Exception('Maximum depth {} reached'.format(max_depth))
			else:
				v_new = get_json_dumpable(
					data=v.copy(),
					depth=depth+1,
					max_depth=max_depth
				)
		else:
			raise TypeError('Unsupported type {}'.format(type(v)))

		if isinstance(k, (str, bool, pandas.datetime)):
			k_new = str(k)
		elif isinstance(k, float_types):
			k_new = float(k)
		elif isinstance(k, int_types):
			k_new = int(k)
		else:
			raise TypeError('Unsupported type {}'.format(type(k)))

		prc_data.update({k_new: v_new})

	return prc_data


def jprint(j, indent=2):
	print(json.dumps(get_json_dumpable(j), indent=indent))


def describe(
		df: pandas.DataFrame, columns=None, excluded_columns=None,
		n_max_levels=100, n_bins=100):
	"""
	:param df: pandas.DataFrame to plot
	:param columns: columns to plot. None plots all the columns in df
	:param excluded_columns: columns to exclude
	:param n_max_levels: maximum number of levels to plot. Variables with more
			than n_max_levels will not be plot
	:param n_bins: nunber of bins for float/int variables
	"""
	original_figsize = plt.rcParams['figure.figsize']
	plt.rcParams['figure.figsize'] = (18, original_figsize[1])

	if columns is None:
		columns = df.columns
	if excluded_columns is not None:
		columns = [x for x in columns if x not in excluded_columns]

	for col in columns:
		dtype = df[col].dtype
		uniques = df[col].unique()
		n_uniques = len(uniques)

		infos = '[variable: {}] [dtype: {}] [n_uniques: {}] '\
			.format(col, dtype, n_uniques)
		print('{:_<125} '.format(infos))

		if dtype in float_types:
			df[col].hist(bins=n_bins)
			plt.tick_params('x', rotation=45)
		elif dtype in int_types:
			if n_uniques <= n_max_levels:
				sns.countplot(df[col])
				plt.tick_params('x', rotation=45)
			else:
				df[col].hist(bins=n_bins)
				plt.tick_params('x', rotation=45)
		elif dtype == 'O' and n_uniques < n_max_levels:
			sns.countplot(df[col])
			plt.tick_params('x', rotation=45)

		plt.show()

	plt.rcParams['figure.figsize'] = original_figsize


# auxiliary functions
def inv_logit(x):
	return numpy.exp(x) / (1 + numpy.exp(x))


def linear(x):
	return x


def reciprocal(x):
	return 1.0 / x


def logit(x):
	return numpy.log(x / (1 - x))



import numpy
import pandas
import sklearn
from sklearn.feature_extraction import FeatureHasher


class Encoder:
    def __init__(self, fit_variables: list, verbose: bool = False):
        self._fit_variables = fit_variables
        self._new_variables = {}
        self._verbose = verbose

        self._fitted = False

        self._is_df = False
        self._is_series = False
        self._is_dict = False
        self._is_array = False

    def reset(self):
        self._fitted = False
        self._new_variables = {}
        self._is_df = False
        self._is_series = False
        self._is_dict = False
        self._is_array = False

    def base_transform(self, data, array_indices, get_transformed):
        if not self._fitted:
            raise FitError('Encoder not fitted')

        self._is_df = isinstance(data, pandas.DataFrame)
        self._is_series = isinstance(data, pandas.Series)
        self._is_dict = isinstance(data, dict)
        self._is_array = isinstance(data, numpy.ndarray)

        if self._is_df:
            skip_data = data[data.columns.difference(self._fit_variables)]
            original_columns = list(data.columns)
            transformed_data = pandas.DataFrame(index=data.index)
        elif self._is_series:
            skip_data = data[data.index.difference(self._fit_variables)]
            original_columns = list(data.index)
            transformed_data = pandas.Series(name=data.name)
        elif self._is_dict:
            skip_data = None
            original_columns = list(data.keys())
            transformed_data = data.copy()
        elif self._is_array:
            skip_data = None
            original_columns = None
            transformed_data = numpy.array([])
            if array_indices is None:
                raise MissingIndicesError(
                    'Missing indices for array transformation'
                )
        else:
            raise TypeError('Type not valid for transformation')

        final_columns = []
        if self._is_series or self._is_df or self._is_dict:
            for variable in original_columns:
                if variable in self._fit_variables:
                    if self._verbose:
                        print('Transforming', variable)

                    transformed_col_or_val, new_columns = get_transformed(
                        col_or_val=data[variable],
                        variable=variable
                    )
                    final_columns += new_columns

                    if self._is_series:
                        transformed_data = transformed_data\
                            .append(transformed_col_or_val)
                    elif self._is_df:
                        transformed_data = transformed_data\
                            .join(transformed_col_or_val)
                    elif self._is_dict:
                        transformed_data.pop(variable)
                        transformed_data.update(transformed_col_or_val)
                else:
                    final_columns.append(variable)

        elif self._is_array:
            index = -1
            for i in range(len(data)):
                if i in array_indices:
                    if self._verbose:
                        print('Transforming', i)

                    index += 1
                    variable = self._fit_variables[index]
                    transformed_col_or_val, new_columns = get_transformed(
                        col_or_val=data[i], variable=variable
                    )
                    transformed_data = numpy.append(
                        transformed_data, transformed_col_or_val
                    )
                else:
                    transformed_data = numpy.append(transformed_data, data[i])

        if self._is_series:
            result = skip_data.append(transformed_data)[final_columns]
        elif self._is_df:
            result = skip_data.join(transformed_data)[final_columns]
        elif self._is_dict:
            result = transformed_data
        else:
            result = transformed_data

        return result

    def fit(self, df: pandas.DataFrame, return_new_variables: bool = False):
        pass


class HashEncoder(Encoder):
    def __init__(self, fit_variables: list, n_min_features: int = 2,
                 n_attempt_max: int = 3, n_attempt_tot_max: int = 30,
                 verbose: bool = False):

        super().__init__(fit_variables, verbose)

        self._hashing_dict = {}
        self._n_min_features = n_min_features
        self._n_attempt_max = n_attempt_max
        self._n_attempt_tot_max = n_attempt_tot_max

    def fit(self, df: pandas.DataFrame, return_new_variables: bool = False):
        self.reset()
        self._hashing_dict = {}

        for variable in self._fit_variables:
            if self._verbose:
                print('Fitting', variable)
            self.__get_hashing(df, variable)

        self._fitted = True
        if return_new_variables:
            return self._new_variables

    def transform(self, data, array_indices=None):
        return super().base_transform(
            data=data,
            array_indices=array_indices,
            get_transformed=self.__get_transformed
        )

    def fit_transform(self, df, return_hashing_cols: bool = False,
                      array_indices=None):
        self.fit(df=df, return_new_variables=return_hashing_cols)
        return self.transform(data=df, array_indices=array_indices)

    def __get_transformed(self, col_or_val, variable: str):
        new_columns = self._new_variables[variable]

        if self._is_series:
            transformed_col_or_val = pandas.Series(
                self._hashing_dict[variable][col_or_val], index=new_columns
            )
        elif self._is_df:
            transformed_col_or_val = pandas.DataFrame(
                data=list(col_or_val.apply(
                    lambda x: self._hashing_dict[variable][x])
                ),
                index=col_or_val.index, columns=new_columns)
        elif self._is_dict:
            try:
                hashing_values = self._hashing_dict[variable][col_or_val]
            except KeyError:
                raise HashingError(
                    'Value \'{}\' not found in hashing dict for '
                    'variable \'{}\'.'.format(col_or_val, variable))

            transformed_col_or_val = {
                new_columns[i]: float(v) for i, v in enumerate(hashing_values)}
        else:
            transformed_col_or_val = numpy.array(
                self._hashing_dict[variable][col_or_val])

        return transformed_col_or_val, new_columns

    def __get_hashing(self, df: pandas.DataFrame, variable: str):
        n_features = self._n_min_features
        n_attempt = 0
        n_attempt_tot = 0
        lost = -1
        non_trivial = False
        uniques = df.loc[:, variable].unique()
        hashed_matrix = [[]]

        while (lost != 0 or not non_trivial) \
                and n_attempt_tot < self._n_attempt_tot_max:
            non_trivial = False
            feature_hasher = FeatureHasher(
                input_type='string',
                n_features=n_features)
            keys_list = list(
                map(lambda x: str(x) + str(numpy.random.rand())[2:8], uniques))

            hashed_matrix = feature_hasher.fit_transform(keys_list)\
                .toarray().astype(int)
            hashed_code = set(
                map(lambda x: ''.join(x.astype(str)), hashed_matrix))

            lost = len(hashed_code) - len(uniques)

            n_attempt += 1
            n_attempt_tot += 1

            if n_attempt == self._n_attempt_max:
                n_attempt = 0
                n_features += 1

            if min(numpy.var(hashed_matrix, axis=0)) > 0:
                non_trivial = True

        if lost != 0 or not non_trivial:
            raise HashingError('Hashing failed on variable: {}. Lost: {}. '
                               'Trivial: {}.'
                               .format(variable, lost, not non_trivial))

        else:
            tmp_hashing_dict = dict(zip(uniques, map(list, hashed_matrix)))
            self._hashing_dict.update({variable: tmp_hashing_dict})
            self._new_variables.update({
                variable: ['{}_H{:03}'.format(variable, i)
                           for i in range(hashed_matrix.shape[1])]
            })


class CircularEncoder(Encoder):
    def __init__(self, fit_variables: list, verbose: bool):
        super().__init__(fit_variables=fit_variables, verbose=verbose)
        self._variables = fit_variables
        self._parameters = {}

    def fit(self, df: pandas.DataFrame, ranges: dict = None):
        self.reset()
        self._parameters = {}

        for variable in self._variables:
            if ranges is None:
                min_value = df[variable].min()
                max_value = df[variable].max()
            else:
                if variable in ranges.keys():
                    min_value = ranges[0]
                    max_value = ranges[1]
                else:
                    raise Exception(
                        'Error! dict \'ranges\' dosn\'t '
                        'contain variable {}'.format(variable))

            self._parameters.update({variable: {
                'min': min_value, 'den': max_value - min_value + 1}})
            self._new_variables.update({
                variable: ['{}_{}'.format(variable, s) for s in ['C', 'S']]})

        self._fitted = True

    def transform(self, data, array_indices=None):
        return super().base_transform(
            data=data,
            array_indices=array_indices,
            get_transformed=self.__get_transformed)

    def fit_transform(self, df, ranges: dict = None, array_indices=None):
        self.fit(df=df, ranges=ranges)
        return self.transform(data=df, array_indices=array_indices)

    def __get_transformed(self, col_or_val, variable: str):
        new_columns = self._new_variables[variable]
        if self._is_df:
            data = list(col_or_val.apply(
                lambda x: self.__get_mapped(x, variable)))
            transformed_col_or_val = pandas.DataFrame(
                data=data,
                index=col_or_val.index,
                columns=new_columns)
        else:
            mapped_values = self.__get_mapped(col_or_val, variable)

            if self._is_series:
                transformed_col_or_val = pandas.Series(
                    data=mapped_values, index=new_columns)
            elif self._is_dict:
                transformed_col_or_val = {
                    new_columns[i]: float(v)
                    for i, v in enumerate(mapped_values)}
            else:
                transformed_col_or_val = numpy.array(mapped_values)

        return transformed_col_or_val, new_columns

    def __get_mapped(self, x, variable):
        theta = 2 * numpy.pi * (x - self._parameters[variable]['min']) / \
                (self._parameters[variable]['den'])
        return [numpy.cos(theta), numpy.sin(theta)]


class OneHotEncoder(Encoder):
    def __init__(self, fit_variables, verbose: bool = False, **kwargs):
        """
        :param fit_variables: names of the variables to use
        :param verbose: print log while processing
        :param kwargs: kwargs to pass to sklearn.preprocessing.OneHotEncoder
        """
        super().__init__(fit_variables=fit_variables, verbose=verbose)
        self._variables = fit_variables

        ohe_args = {'categories': 'auto', 'sparse': False}
        ohe_args.update(kwargs)

        self.__one_hot_encoders = {
            variable: sklearn.preprocessing.OneHotEncoder(**ohe_args)
            for variable in self._variables}

    def fit(self, df: pandas.DataFrame, return_new_variables: bool = False):
        self.reset()
        for variable, encoder in self.__one_hot_encoders.items():
            encoder.fit(df[[variable]])

        self._new_variables = {
            variable: ['{}_OHE{:03}'.format(variable, i)
                       for i, _ in enumerate(df[variable].unique())]
            for variable in self._variables}

        self._fitted = True
        if return_new_variables:
            return self._new_variables

    def transform(self, data, array_indices=None):
        return super().base_transform(
            data=data,
            array_indices=array_indices,
            get_transformed=self.__get_transformed
        )

    def fit_transform(self, df, array_indices=None):
        self.fit(df=df)
        return self.transform(data=df, array_indices=array_indices)

    def __get_transformed(self, col_or_val, variable: str):
        new_columns = self._new_variables[variable]

        if self._is_df:
            input_col_or_val = col_or_val.values.reshape(-1, 1)
            data = self.__one_hot_encoders[variable].transform(input_col_or_val)
            transformed_col_or_val = pandas.DataFrame(
                data=data,
                index=col_or_val.index,
                columns=new_columns
            )
        else:
            data = self.__one_hot_encoders[variable].transform([[col_or_val]])
            data = data[0]

            if self._is_series:
                transformed_col_or_val = pandas.Series(
                    data=data, index=new_columns)
            elif self._is_dict:
                transformed_col_or_val = {
                    new_columns[i]: float(v) for i, v in enumerate(data)}
            else:
                transformed_col_or_val = data

        return transformed_col_or_val, new_columns


class KBinsDiscretizer(Encoder):
    def __init__(self, fit_variables, verbose: bool = False, **kwargs):
        """
        :param fit_variables: names of the variables to use
        :param verbose: print log while processing
        :param kwargs: kwargs to pass to sklearn.preprocessing.OneHotEncoder
        """
        super().__init__(fit_variables=fit_variables, verbose=verbose)
        self._variables = fit_variables

        self._k_bins_discretizers = {
            variable: sklearn.preprocessing.KBinsDiscretizer(**kwargs)
            for variable in self._variables
        }

    def fit(self, df: pandas.DataFrame, return_new_variables: bool = False):
        self.reset()

        for variable, encoder in self._k_bins_discretizers.items():
            encoder.fit(df[[variable]])
            new_variables = [
                '{}_KBD{:03}'.format(variable, i)
                for i in range(encoder.n_bins_[0])]

            self._new_variables.update({variable: new_variables})

        self._fitted = True
        if return_new_variables:
            return self._new_variables

    def transform(self, data, array_indices=None):
        return super().base_transform(
            data=data,
            array_indices=array_indices,
            get_transformed=self.__get_transformed
        )

    def fit_transform(self, df, array_indices=None):
        self.fit(df=df)
        return self.transform(data=df, array_indices=array_indices)

    def __get_transformed(self, col_or_val, variable: str):
        new_columns = self._new_variables[variable]
        print(new_columns)
        if self._is_df:
            input_col_or_val = col_or_val.values.reshape(-1, 1)
            data = self._k_bins_discretizers[variable].transform(
                data=input_col_or_val)
            transformed_col_or_val = pandas.DataFrame(
                data=data,
                index=col_or_val.index,
                columns=new_columns
            )
        else:
            data = self._k_bins_discretizers[variable].transform([[col_or_val]])
            data = data[0]

            if self._is_series:
                transformed_col_or_val = pandas.Series(
                    data=data, index=new_columns)
            elif self._is_dict:
                transformed_col_or_val = {
                    new_columns[i]: float(v) for i, v in enumerate(data)}
            else:
                transformed_col_or_val = data

        return transformed_col_or_val, new_columns

#
# class KBinsPowers(KBinsDiscretizer):
# 	def __init__(self, variables, verbose: bool = False, **kwargs):
# 		"""
# 		:param variables:
# 		:param verbose:
# 		:param kwargs:
# 		{
# 			'power' : <int>
# 			'variables': [ <variables list> ]
# 		}
# 		{
# 			'power' : <int>
# 			'variables': None
# 		}
# 		{
# 			'powers' :{
# 				variable : (None, [powers])
# 			}
# 		}
# 		{
# 			'powers' :{
# 				variable : ([ <bin indeces> ], [powers])
# 			}
# 		}
# 		"""
# 		super().__init__(variables=variables, verbose=verbose, **kwargs)
#
# 	def fit(self, df: pandas.DataFrame, **kwargs):
# 		super().fit(df, **kwargs)
# 		self._fitted = False
#
# 		power = kwargs.get('power')
# 		if power is not None:
# 			variables = kwargs.get('variables')
# 			if variables is None:
# 				variables = self._variables
#
# 			for variable in variables:
# 				dx = self.__get_dx(s=df[variable], variable=variable)
#
# 		elif 'powers' in kwargs.keys():
# 			for variable, powers in kwargs['powers'].items():
# 				encoder = self._k_bins_discretizers[variable]
#
# 	# encoder.fit(df[[variable]])
# 	# self.__new_columns.update({variable: ['{}_KBD{:03}'.
# 	format(variable, i) for i in range(encoder.n_bins_[0])]})
# 	def __get_dx(self, s: pandas.Series, variable: str):
# 		x0 = self._k_bins_discretizers[variable].bin_edges_
# 		dx = self._k_bins_discretizers[variable].transform(s)
# 		return dx


class LabelEncoder(Encoder):
    def __init__(self, fit_variables, verbose=False):
        super().__init__(fit_variables=fit_variables, verbose=verbose)
        self.__label_encoders = {
            variable: sklearn.preprocessing.LabelEncoder()
            for variable in self._fit_variables}

    def fit(self, df: pandas.DataFrame, return_new_variables: bool = False):
        self.reset()

        for variable, encoder in self.__label_encoders.items():
            encoder.fit(df[[variable]])

        self._fitted = True
        if return_new_variables:
            return self._new_variables

    def transform(self, data, array_indices=None):
        return super().base_transform(
            data=data,
            array_indices=array_indices,
            get_transformed=self.__get_transformed)

    def fit_transform(self, df, array_indices=None):
        self.fit(df=df)
        return self.transform(data=df, array_indices=array_indices)

    def __get_transformed(self, col_or_val, variable: str):
        if self._is_df:
            transformed_col_or_val = pandas.DataFrame(
                self.__label_encoders[variable].transform(col_or_val.values)
                    .reshape(-1, 1), index=col_or_val.index, columns=[variable])
        else:
            data = self.__label_encoders[variable].transform([col_or_val])[0]
            if self._is_series:
                transformed_col_or_val = pandas.Series(
                    data=data, index=[variable])
            elif self._is_dict:
                transformed_col_or_val = {variable: data}
            else:
                transformed_col_or_val = data

        return transformed_col_or_val, variable


class SequentialEncoder:
    def __init__(self, transformers: list):
        """
        :param transformers: list of tuples with shape:
            ('encoder_name', encoder_class, <encoder_init_parameters>)
        """
        self.__encoders = {
            name: encoder(**kwargs) for name, encoder, kwargs in transformers}

    def fit(self, df: pandas.DataFrame, fit_parameters: dict = None):
        """
        :param df: pandas.DataFrame to fit
        :param fit_parameters: parameters to pass to fit functions of each
            encoder: {'encoder_name_1' :
                            {'variable_1': value_1, 'variable_2': value_2} ... }
        """
        if fit_parameters is not None:
            fit_parameters.update({
                name: {}
                for name in self.__encoders.keys()
                if name not in fit_parameters.keys()})
        else:
            fit_parameters = {name: {} for name in self.__encoders.keys()}
        for name, encoder in self.__encoders.items():
            encoder.fit(df, **fit_parameters[name])

    def transform(self, data, array_indices: dict = None):
        """
        :param data: data to tranform
        :param array_indices: dict containing indices to use in case data is a
            numpy.array: {
                'encoder_name_1: [i_1, i_2, ...], 'encoder_name_2: [...]}
        :return:
        """
        if array_indices is not None:
            array_indices.update({
                name: {}
                for name in self.__encoders.keys()
                if name not in array_indices.keys()})
        else:
            array_indices = {name: {} for name in self.__encoders.keys()}

        transformed_data = data.copy()
        for name, encoder in self.__encoders.items():
            transformed_data = encoder.transform(
                data=transformed_data, array_indices=array_indices[name])

        return transformed_data

    def fit_transform(self, df: pandas.DataFrame, fit_parameters: dict = None,
                      array_indices: dict = None):
        """
        :param df: pandas.DataFrame to fit
        :param fit_parameters: parameters to pass to fit functions of each
            encoder: {'encoder_name_1' :
                {'variable_1': value_1, 'variable_2': value_2} ... }
        :param array_indices: dict containing indices to use in case data is a
            numpy.array:{
                'encoder_name_1: [i_1, i_2, ...], 'encoder_name_2: [...]}
        :return:
        """
        self.fit(df=df, fit_parameters=fit_parameters)
        return self.transform(data=df, array_indices=array_indices)


# EXCEPTIONS -------------------------------------------------------------------
class FitError(Exception):
    """:raises Error when class Hasher is not fitted"""


class HashingError(Exception):
    """:raises Error when hashing is invalid"""


class MissingIndicesError(Exception):
    """:raises Error when indices are missing"""



import pandas


class DataReader:
    def __init__(self, default_values_path):
        self._process = None

        self._input_data = {}
        self._input_data_processed = {}
        self._suppress_calculation = False

        self._warnings = []
        self._errors = []

        # default values -------------------------------------------------------
        self._variables_set_to_default = []
        with open(default_values_path, 'r') as f:
            lines = [x.replace('\n', '').split(',') for x in f.readlines()[1:]]
            self._default_values = {k: v for (k, v) in lines}

        for variable, default_value in self._default_values.items():
            try:
                new_default_value = float(default_value)
            except ValueError:
                pass
            else:
                if int(new_default_value) == new_default_value:
                    new_default_value = int(new_default_value)
                self._default_values.update({variable: new_default_value})

    def _message_key_error(
            self,
            error: KeyError,
            additional_message: str = ''
    ) -> str:
        return 'KeyError ({}): missing key {} in input_data dict. ' \
               'Setting default value.{}' \
            .format(self._process, error, additional_message)

    def _message_key_error_crucial(
            self,
            error: KeyError) -> str:
        return 'KeyError ({}): missing crucial key {} in input_data dict. ' \
               'Calculation suppressed.'.format(self._process, error)

    def _message_value_error(
            self,
            variable: str,
            wrong_value,
            additional_message: str = ''
    ) -> str:
        return 'ValueError ({}): wrong value \'{}\' for variable \'{}\'. ' \
               'Setting default value. {}' \
            .format(self._process, wrong_value, variable, additional_message)

    def _message_type_error(
            self,
            variable: str,
            wrong_value,
            additional_message: str = ''
    ) -> str:
        return 'TypeError ({}): wrong type \'{}\' for variable \'{}\'. ' \
               'Setting default value. {}' \
            .format(self._process, type(wrong_value),
                    variable, additional_message)

    def _message_choice_error(
            self,
            variable,
            wrong_value,
            additional_message: str = ''
    ):
        return 'ChoiceError ({}): value \'{}\' not found in possible choices ' \
               'for variable \'{}\'. Setting default value. {}' \
            .format(self._process, wrong_value, variable, additional_message)

    def _message_attribute_error(
            self,
            variable, error,
            additional_message: str = ''
    ):
        return 'AttributeError ({}): {}. Variable \'{}\'. ' \
               'Setting default value. {}' \
            .format(self._process, error, variable, additional_message)

    def _handle_exceptions(
            self,
            exceptions_handler: list,
            variable: str,
            error: Exception,
            wrong_value,
            additional_message: str = None
    ):
        if isinstance(error, ValueError):
            exceptions_handler.append(self._message_value_error(
                variable=variable,
                wrong_value=wrong_value,
                additional_message=additional_message))
        elif isinstance(error, TypeError):
            exceptions_handler.append(self._message_type_error(
                variable=variable,
                wrong_value=wrong_value,
                additional_message=additional_message))
        else:
            exceptions_handler.append('Unhandled exception thrown: {}'
                                      .format(error))

    def _process_str(self, value, exceptions_handler,
                     set_lowercase: bool = False, set_uppercase: bool = False,
                     force_type=None):
        is_error = False

        if set_lowercase:
            value = str(value).lower()
        if set_uppercase:
            value = str(value).upper()
        if force_type is not None:
            try:
                value = force_type(value)
            except ValueError:
                is_error = True
                exceptions_handler.append(
                    'ValueError ({}): impossible to force type '
                    '{} on {}.'.format(self._process, force_type, value)
                )
        return value, is_error

    def _get_value(
            self,
            variable: str,
            exceptions_handler: list,
            is_crucial: bool = False,
            get_error: bool = False,
            new_name=None,
            set_default=True
    ):
        name = variable if new_name is None else new_name
        is_error = False
        value = None

        try:
            value = self._input_data[variable]
        except KeyError as kE:
            is_error = True

            if is_crucial:
                exceptions_handler.append(
                    self._message_key_error_crucial(error=kE)
                )
                self._suppress_calculation = True
            else:
                if set_default:
                    value = self._set_default_value(name)
                exceptions_handler.append(
                    self._message_key_error(error=kE)
                )

        if get_error:
            return value, is_error
        else:
            return value

    def _get_str(
            self,
            variable: str,
            exceptions_handler: list,
            new_name: str = None,
            set_lowercase: bool = False,
            set_uppercase: bool = False,
            is_crucial: bool = False
    ):
        name = variable if new_name is None else new_name
        value, is_error = self._get_value(
            variable=variable,
            exceptions_handler=exceptions_handler,
            is_crucial=is_crucial,
            get_error=True,
            new_name=new_name
        )

        if not is_error:
            try:
                value = str(value)
            except (ValueError, TypeError) as error:
                self._handle_exceptions(
                    exceptions_handler=exceptions_handler,
                    variable=variable,
                    error=error,
                    wrong_value=value
                )

                if is_crucial:
                    value = None
                    self._suppress_calculation = True
                else:
                    value = self._set_default_value(name)
            else:
                if set_lowercase:
                    value = value.lower()
                if set_uppercase:
                    value = str(value).upper()

        return value

    def _set_str(
            self,
            variable: str,
            exceptions_handler: list,
            new_name: str = None,
            set_lowercase: bool = False,
            set_uppercase: bool = False,
            is_crucial: bool = False
    ):
        name = variable if new_name is None else new_name
        value = self._get_str(
            variable=variable,
            exceptions_handler=exceptions_handler,
            new_name=new_name,
            set_lowercase=set_lowercase,
            set_uppercase=set_uppercase,
            is_crucial=is_crucial
        )
        self._input_data_processed.update({name: value})

    def _get_categorical(
            self,
            variable: str,
            choices: list,
            exceptions_handler: list,
            force_type=None,
            new_name: str = None,
            set_lowercase: bool = False,
            set_uppercase: bool = False,
            is_crucial: bool = False
    ):
        name = variable if new_name is None else new_name
        value, is_error = self._get_value(
            variable=variable,
            exceptions_handler=exceptions_handler,
            is_crucial=is_crucial,
            get_error=True,
            new_name=new_name
        )

        if not is_error:
            value, is_error = self._process_str(
                value=value,
                exceptions_handler=exceptions_handler,
                set_lowercase=set_lowercase,
                set_uppercase=set_uppercase,
                force_type=force_type
            )

            if value not in choices:
                is_error = True
                exceptions_handler.append(
                    self._message_choice_error(variable=name, wrong_value=value)
                )

            if is_error:
                if is_crucial:
                    value = None
                    self._suppress_calculation = True
                else:
                    value = self._set_default_value(name)

        return value

    def _set_categorical(
            self,
            variable: str,
            choices: list,
            exceptions_handler: list,
            force_type=None,
            new_name: str = None,
            set_lowercase: bool = False,
            set_uppercase: bool = False,
            is_crucial: bool = False
    ):
        name = variable if new_name is None else new_name
        value = self._get_categorical(
            variable=variable,
            choices=choices,
            exceptions_handler=exceptions_handler,
            force_type=force_type,
            new_name=new_name,
            set_lowercase=set_lowercase,
            set_uppercase=set_uppercase,
            is_crucial=is_crucial
        )
        self._input_data_processed.update({name: value})

    def _get_numeric(
            self,
            variable: str,
            exceptions_handler: list,
            numeric_type: type = int,
            min_value: int = None,
            max_value: int = None,
            bucket: int = None,
            new_name: str = None,
            is_crucial=False
    ):
        name = variable if new_name is None else new_name
        value, is_error = self._get_value(
            variable=variable,
            exceptions_handler=exceptions_handler,
            is_crucial=is_crucial,
            get_error=True,
            new_name=new_name
        )

        if not is_error:
            try:
                value = numeric_type(value)
            except (ValueError, TypeError) as error:
                self._handle_exceptions(
                    exceptions_handler=exceptions_handler,
                    variable=name,
                    error=error,
                    wrong_value=value
                )
                if is_crucial:
                    value = None
                    self._suppress_calculation = True
                else:
                    value = self._set_default_value(name)
            else:
                if value != bucket:
                    if min_value is not None:
                        value = max(min_value, value)
                    if max_value is not None:
                        value = min(max_value, value)

        return value

    def _set_numeric(
            self,
            variable: str,
            exceptions_handler: list,
            numeric_type: type = int,
            min_value: int = None,
            max_value: int = None,
            bucket: int = None,
            new_name: str = None,
            is_crucial=False
    ):
        name = variable if new_name is None else new_name
        value = self._get_numeric(
            variable=variable,
            exceptions_handler=exceptions_handler,
            numeric_type=numeric_type,
            min_value=min_value,
            max_value=max_value,
            bucket=bucket,
            new_name=new_name,
            is_crucial=is_crucial
        )
        self._input_data_processed.update({name: value})

    def _get_date(
            self,
            variable: str,
            exceptions_handler: list,
            new_name: str = None,
            is_crucial: bool = False,
            timezone: str = 'Europe/London'
    ):
        name = variable if new_name is None else new_name
        date, is_error = self._get_value(
            variable=variable,
            exceptions_handler=exceptions_handler,
            is_crucial=is_crucial,
            get_error=True,
            new_name=new_name
        )

        if not is_error:
            try:
                date = pandas.to_datetime(date)
                if date is None:
                    raise ValueError
            except (ValueError, AttributeError) as error:
                if isinstance(error, ValueError):
                    self._warnings.append(
                        self._message_value_error(
                            variable=name,
                            wrong_value=date
                        )
                    )
                if isinstance(error, AttributeError):
                    self._warnings.append(
                        self._message_attribute_error(
                            variable=name,
                            error=date
                        )
                    )

                if is_crucial:
                    date = None
                    self._suppress_calculation = True
                else:
                    date = self._set_default_value(name)
            else:
                date = date.tz_convert(timezone)

        return date

    def _set_date(
            self,
            variable: str,
            exceptions_handler: list,
            new_name: str = None,
            is_crucial=False,
            timezone: str = 'Europe/London'
    ):
        name = variable if new_name is None else new_name
        date = self._get_date(
            variable=variable,
            exceptions_handler=exceptions_handler,
            new_name=new_name,
            is_crucial=is_crucial,
            timezone=timezone
        )
        self._input_data_processed.update({name: date})

    def _set_default_value(self, variable):
        try:
            default_value = self._default_values[variable]
        except KeyError:
            raise KeyError(
                'Missing default value for variable {}'.format(variable)
            )
        else:
            self._variables_set_to_default.append(variable)

        return default_value

    def _reset(self):
        self._process = None

        self._input_data = {}
        self._input_data_processed = {}
        self._variables_set_to_default = []

        self._warnings = []
        self._errors = []

        self._suppress_calculation = False



import pandas
import numpy
import seaborn

from matplotlib import pyplot as plt
from itertools import groupby
from sklearn.cluster import SpectralClustering


def norm_cov_matrix(df):
	"""
	:param df: pandas dataframe to use
	:return: normalized covariance matrix of a pandas dataframe
	"""
	var = numpy.sqrt(df.var())
	return (df.loc[:, var.index]/var).cov()


def cluster_variables(df: pandas.DataFrame, n_clusters: int, sigma: float, selected_variables: list=None, target: list=None, return_results: bool=False, **kwargs):
	"""
	:param df: pandas dataframe
	:param n_clusters: number of cluster to create
	:param sigma: decay constant for affinity matrix
	:param selected_variables: varaibles to use in clustering
	:param target: optional target variable to spot correlations
	:param return_results: return full clusters description
	:param kwargs: kwargs for seaborn heatmap
	:return: full clusters description
	"""
	# Affinity ------------------------------------------------------------------------------
	if selected_variables is None:
		cov_matrix = numpy.abs(norm_cov_matrix(df))
	else:
		if target[0] is not None:
			columns = selected_variables
			for tgt in target:
				if tgt not in selected_variables:
					columns = columns + [tgt]
		else:
			columns = selected_variables

		cov_matrix = numpy.abs(norm_cov_matrix(df[columns]))

	columns = list(cov_matrix.columns)
	selected_variables = columns

	for tgt in target:
		if tgt in selected_variables:
			selected_variables.remove(tgt)

	dist_matrix = 1. - cov_matrix

	affinity_matrix = numpy.exp(- dist_matrix ** 2 / (2. * sigma ** 2))

	# Clusters ------------------------------------------------------------------------------
	cluster = SpectralClustering(affinity='precomputed', n_clusters=n_clusters)
	clusters_tags = cluster.fit_predict(affinity_matrix.loc[selected_variables, :].loc[:, selected_variables])
	ordered_variables = [x for _, x in sorted(zip(clusters_tags, selected_variables))]

	clusters = {}
	for key, value in sorted(dict(zip(selected_variables, clusters_tags)).items()):
		clusters.setdefault(value, []).append(key)

	clusters_dim = [len(list(group)) for key, group in groupby(sorted(clusters_tags))]

	if target[0] is not None:
		columns = ordered_variables + target
	else:
		columns = ordered_variables

	sorted_cov_matrix = cov_matrix.loc[columns, :].loc[:, columns]
	sorted_affinity_matrix = affinity_matrix.loc[columns, :].loc[:, columns]

	# Plots ------------------------------------------------------------------------------
	plt.rcParams['figure.figsize'] = (18, max(10, len(cov_matrix.columns)//5))

	titles = ['sorted_cov_matrix', 'sorted_affinity_matrix']
	for i, matrix in enumerate([sorted_cov_matrix, sorted_affinity_matrix]):
		g = seaborn.heatmap(matrix, cmap='viridis', linecolor='black', linewidths=0.1, **kwargs)
		g.axes.hlines(numpy.cumsum(clusters_dim)[:-1], *g.axes.get_xlim(), colors='white', )
		g.axes.vlines(numpy.cumsum(clusters_dim)[:-1], *g.axes.get_xlim(), colors='white')

		if target is not None:
			g.axes.hlines(numpy.cumsum(clusters_dim)[-1], *g.axes.get_xlim(), colors='red', )
			g.axes.vlines(numpy.cumsum(clusters_dim)[-1], *g.axes.get_xlim(), colors='red')

		g.set_title(titles[i])
		plt.show()

	if return_results:
		return {
			'clusters': clusters,
			'cov_matrix': cov_matrix,
			'affinity_matrix': affinity_matrix,
			'sorted_cov_matrix': sorted_cov_matrix,
			'sorted_affinity_matrix': sorted_affinity_matrix}



import numpy
import pandas


class Scaler:
	def __init__(self, scaler_class, fit_variables):
		self.__scaler_class = scaler_class
		self.__fit_variables = fit_variables
		self.__scalers = {}
		self.__fitted = False

	def reset(self):
		self.__scalers = {}
		self.__fitted = False

	def fit(self, df: pandas.DataFrame):
		self.reset()

		for variable in self.__fit_variables:
			scaler = self.__scaler_class()
			scaler.fit(df[[variable]])
			self.__scalers.update({variable: scaler})

		self.__fitted = True

	def transform(self, data, transform_variables: list = None, array_indices: list = None):
		if transform_variables is None:
			transform_variables = self.__fit_variables
		if self.__fitted:
			# transformed = None
			if isinstance(data, dict):
				transformed = data.copy()
				for variable in transform_variables:
					transformed.update({variable: self.__scalers[variable].transform([[data[variable]]])[0, 0]})

			elif isinstance(data, pandas.DataFrame):
				original_variables = list(data.columns)
				other_variables = list(data.columns.difference(transform_variables))
				transformed = pandas.DataFrame(index=data.index)
				for variable in transform_variables:
					transformed = transformed.join(pandas.DataFrame(self.__scalers[variable].transform(data[[variable]]), index=data.index, columns=[variable]))
				transformed = data[other_variables].join(transformed)[original_variables]

			elif isinstance(data, pandas.Series):
				original_variables = list(data.index)
				other_variables = list(data.index.difference(transform_variables))
				transformed = pandas.Series(name=data.name)
				for variable in transform_variables:
					transformed = transformed.append(pandas.Series(self.__scalers[variable].transform([[data[variable]]])[0], index=[variable]))
				transformed = data[other_variables].append(transformed)[original_variables]

			elif isinstance(data, numpy.ndarray):
				if len(data.shape) != 2:
					raise Exception('2D array expected. Got {}D array instead'.format(len(data.shape)))
				if len(array_indices) == len(transform_variables):
					transformed = numpy.array([]).reshape(data.shape[0], 0)
					for i, index in enumerate(range(data.shape[1])):
						if index in array_indices:
							variable = transform_variables[i]
							transformed = numpy.concatenate([transformed, self.__scalers[variable].transform(data[:, [index]])], axis=1)
						else:
							transformed = numpy.concatenate([transformed, data[:, [i]]], axis=1)
				else:
					raise Exception('Wrong number of array_indices! Expected {}, got {}'.format(len(transform_variables), len(array_indices)))

			else:
				raise Exception('Type {} is invalid for scaling transformation'.format(type(data)))

		else:
			raise Exception('Scaler not fitted')

		return transformed

	def fit_transform(self, df: pandas.DataFrame):
		self.fit(df)
		return self.transform(df)


class MultiDimStandardScaler:
	def __init__(self):
		self._mean = None
		self._var = None
		self._std = None
		self.__fitted = None

	def fit(self, df):
		if isinstance(df, pandas.DataFrame):
			data = df.values
		else:
			data = df.copy()

		self._mean = numpy.mean(data, axis=0)
		self._var = numpy.var(data, axis=0)
		self._std = numpy.sqrt(self._var)
		self.__fitted = True

	def transform(self, df):
		if self.__fitted:
			sample_len = len(df)
			scaled = df - numpy.array([self._mean for _ in range(sample_len)])
			scaled = scaled / numpy.array([self._std for _ in range(sample_len)])
		else:
			raise Exception('Scaler not fitted.')

		return scaled

	def fit_transform(self, df):
		self.fit(df)
		return self.transform(df)



from setuptools import setup, find_packages

version = '1.0.20'

setup(
	name='primalytics',
	packages=find_packages(),
	version=version,
	license='MIT',											# Chose a license from here: https://help.github.com/articles/licensing-a-repository
	description='Analytic library for Prima',				# Give a short description about your library
	author='Stefano Rossotti',
	author_email='stefano.rossotti@hotmail.it',
	url='https://github.com/user/reponame',					# Provide either the link to your github or to your website
	download_url='https://github.com/primait/prima-analytics/archive/{}.tar.gz'.format(version),
	keywords=['prima', 'analytics'], 						# Keywords that define your package best
	install_requires=[
		'numpy==1.16.2',
		'pandas==0.24.2',
		'matplotlib==3.0.3',
		'scikit-learn==0.20.3',
		'seaborn',
		'pydotplus'
	],
	classifiers=[
		'Development Status :: 3 - Alpha',      			# Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
		'Intended Audience :: Developers',      			# Define that your audience are developers
		'Topic :: Software Development :: Build Tools',
		'License :: OSI Approved :: MIT License',   		# Again, pick a license
		'Programming Language :: Python :: 3',      		# Specify which pyhton versions that you want to support
		'Programming Language :: Python :: 3.4',
		'Programming Language :: Python :: 3.5',
		'Programming Language :: Python :: 3.6',
		'Programming Language :: Python :: 3.7',
	],
)


import datetime


def get_date(n_days=0):
	now = datetime.datetime.now() + datetime.timedelta(n_days)
	date = {
		'ye': str(now.year),
		'mo': str(100+now.month)[1:],
		'da': str(100+now.day)[1:],
		'ho': str(100+now.hour)[1:],
		'mi': str(100+now.minute)[1:]
	}
	return date


def get_date_ymd(n_days=0):
	date = get_date(n_days)
	return date['ye'] + date['mo'] + date['da']


def get_date_y_m_d(n_days=0):
	date = get_date(n_days)
	return date['ye'] + '-' + date['mo'] + '-' +  date['da']


def get_date_ymdhm(n_days=0):
	date = get_date(n_days)
	return date['ye'] + date['mo'] + date['da'] + '-' + date['ho'] + date['mi']


import pandas
import seaborn
import pydotplus
import numpy

from sklearn.externals.six import StringIO
from IPython.display import Image
from sklearn.tree import export_graphviz
from matplotlib import pyplot as plt
from matplotlib import gridspec


def plot_univariate(df: pandas.DataFrame, x_list: list, y_list: list, y_list_secondary: list = None, ylim: tuple = (None, None), mean_variables: list = None, order=None, hlines: list = None,
					grid: bool = False):
	"""
	:param df: pandas DataFrame to plot
	:param x_list: x variables to plot
	:param y_list: y variables to plot
	:param y_list_secondary: secondary y variables to plot
	:param ylim: y axis limit
	:param mean_variables: variables of which plot the mean
	:param order: sorting for x values, valid values are 'asc', 'desc', '<column_name>'
	:param hlines: list of float. plots horizontal lines
	:param grid: show grid
	:return: None
	"""
	plt.rcParams['figure.figsize'] = (18, 10)
	order_choices = ['desc', 'asc'] + list(df.columns)
	if y_list_secondary is None:
		y_list_secondary = []

	if len(x_list) > 1 and order is not None:
		if not isinstance(order, dict) and order not in order_choices:
			raise Exception('Error! order provided is invalid. Possible choices are: {} or a dict shaped as {{x: order for x in x_list}}'.format(order_choices))
		if isinstance(order, dict):
			if sum([x not in order.keys() for x in x_list]) > 0:
				raise Exception('Error! not all the variables specified in \'x_list\' are present in \'order\' dict')

	for x in x_list:
		n_uniques = len(df[x].unique())

		ax1 = plt.axes()
		ax1.tick_params('x', rotation=90)

		if order is None:
			plt_order = sorted(df[x].unique())
		elif order == 'desc':
			plt_order = df[x].value_counts().sort_values(ascending=False).index
		elif order == 'asc':
			plt_order = df[x].value_counts().sort_values(ascending=True).index
		elif isinstance(order, str):
			if order in df.columns:
				plt_order = list(df.pivot_table(index=x, aggfunc={order: 'mean'}).sort_values(order).index)
			else:
				raise Exception('Error! cannot order by variable \'{}\' as is not present in input dataset'.format(order))
		elif isinstance(order, dict):
			plt_order = order[x]
		elif isinstance(order, list):
			plt_order = order
			if len(x_list) == 1:
				missing_values = set(df[x].unique()).difference(order)
				if len(missing_values) != 0:
					raise Exception('Error! missing values \'{}\' in order list'.format(missing_values))
			else:
				raise Exception('Error! {} plotting variables specified but only one order list provided.'.format(len(x_list)))

		else:
			raise ValueError('{} invalid value for data ordering!'.format(order))

		seaborn.countplot(x=x, data=df, ax=ax1, alpha=0.7, order=plt_order)

		ax2 = ax1.twinx()
		if ylim[0] is not None and ylim[1] is not None:
			ax2.set_ylim(ylim)

		color = seaborn.color_palette("Set1", 8)
		color_index = 0

		for y in y_list:
			seaborn.pointplot(x=x, y=y, data=df, ax=ax2, color=color[color_index], order=plt_order)
			color_index += 1

		if len(y_list_secondary) > 0:
			ax3 = ax1.twinx()
			for y in y_list_secondary:
				seaborn.pointplot(x=x, y=y, data=df, ax=ax3, color=color[color_index], order=plt_order)
				color_index += 1

			ax3.spines['right'].set_position(('outward', 60))
			ax3.tick_params('y', labelsize=15)

		ax1.tick_params('x', labelsize=15)
		ax1.tick_params('y', labelsize=15)
		ax2.tick_params('y', labelsize=15)

		legend = y_list + y_list_secondary

		if hlines is not None:
			if not isinstance(hlines, list):
				hlines = [hlines]
			for hline in hlines:
				plt.hlines(y=hline, xmin=0, xmax=n_uniques, label=str(hline), colors=color[color_index], linestyles='--')
				color_index += 1

			legend += list(map(str, hlines))

		if mean_variables is not None:
			for mean_variable in mean_variables:
				y_mean = df[mean_variable].mean()
				ax2.plot([y_mean for _ in range(n_uniques)], c=color[color_index])
				color_index += 1
			legend += [mv + '_mean' for mv in mean_variables]

		ax2.legend(legend)

		leg = ax2.get_legend()
		color_index = 0
		for lgd in leg.legendHandles:
			lgd.set_color(color[color_index])
			color_index += 1

		ax2.grid(grid)
		plt.show()


def plot_tree(tree, class_names: list, feature_names: list, weighted_colors: float = None, palette_name: str = 'Accent'):
	"""
	Plots a decision tree with all its leaves
	:param tree: sklearn tree object
	:param class_names: names of the final classes
	:param feature_names: name of the features used in the decision tree
	:param weighted_colors: weight to apply to color scales
	:param palette_name: seaborn palette name
	:return: tree image
	"""
	palette = seaborn.color_palette(palette_name, n_colors=tree.n_classes_)
	colors = numpy.array(list(map(lambda x: 256 * numpy.array(x), palette)))

	dot_data = StringIO()
	export_graphviz(
		tree, out_file=dot_data,
		filled=True, rounded=True,
		special_characters=True,
		feature_names=feature_names,
		class_names=class_names)

	graph_string = dot_data.getvalue()

	for leaf_ID, leaf in enumerate(numpy.where(tree.tree_.children_left == -1)[0]):
		graph_string = graph_string.replace('\n{} [label=<'.format(leaf), '\n{} [label=<----- leaf {} -----<br/>'.format(leaf, leaf_ID))

	graph = pydotplus.graph_from_dot_data(graph_string)

	nodes = graph.get_node_list()

	for node in nodes:
		if node.get_label():
			values = numpy.array([int(ii) for ii in node.get_label().split('value = [')[1].split(']')[0].split(',')])

			if weighted_colors is None:
				color_weights = numpy.zeros(tree.n_classes_)
				color_weights[numpy.where(values == max(values))] = 1
			else:
				color_weights = values / sum(values)
				color_weights = numpy.power(color_weights, weighted_colors) / sum(numpy.power(color_weights, weighted_colors))

			color_rgb = numpy.dot(color_weights, colors)
			color_hex = '#%02x%02x%02x' % tuple(color_rgb.astype(int))
			node.set_fillcolor(color_hex)

	return Image(graph.create_png())


def plot_scatter_hist(df: pandas.DataFrame, x: str, y: str, fsize=(18, 8), width_ratios: tuple = (5, 1), height_ratios: tuple = (4, 1), hue: str = None, s=0.1, bins=100, alpha=0.5):
	"""
	Plots a scatter prlot rounded by an histogram for each axis
	:param df: pandas DataFrame to use
	:param x:
	:param y:
	:param fsize:
	:param width_ratios:
	:param height_ratios:
	:param hue:
	:param s:
	:param bins:
	:param alpha:
	:return:
	"""
	plt.rcParams['figure.figsize'] = fsize
	gs = gridspec.GridSpec(2, 2, width_ratios=list(width_ratios), height_ratios=list(height_ratios))

	ax0 = plt.subplot(gs[0])
	ax1 = plt.subplot(gs[1])
	ax2 = plt.subplot(gs[2])

	if hue is not None:
		legend = []
		for value in df[hue].value_counts().index.values:
			legend.append(value)
			ax0.scatter(df.loc[df[hue] == value, x], df.loc[df[hue] == value, y], s=s)
			ax1.hist(df.loc[df[hue] == value, y], orientation='horizontal', bins=bins, alpha=alpha)
			ax2.hist(df.loc[df[hue] == value, x], orientation='vertical', bins=bins, alpha=alpha)

		ax0.set(xlabel=x, ylabel=y)
		ax1.set(xlabel='count', ylabel=y)
		ax2.set(xlabel=x, ylabel='count')

		ax0.legend(legend, markerscale=6)

	else:
		ax0.scatter(df.loc[:, x], df.loc[:, y], s=s)
		plt.xlabel(x)
		plt.ylabel(y)
		ax1.hist(df.loc[:, y], orientation='horizontal', bins=bins, alpha=alpha)
		ax2.hist(df.loc[:, x], orientation='vertical', bins=bins, alpha=alpha)

	plt.tight_layout()
	plt.show()


def plot_heatmap(df: pandas.DataFrame, var_x: str, var_y: str, var_z: str, z_aggfunc, var_x_n_quantiles: int = 20, var_y_n_quantiles: int = 20, get_bins: bool = False, **kwargs):
	"""
	:param df: pandas DataFrame to use
	:param var_x: first variable for the heatmap
	:param var_y: second variable for the heatmap
	:param var_z: plotting variable for the heatmap
	:param z_aggfunc: aggregating function to use in the heatmap
	:param var_x_n_quantiles:
	:param var_y_n_quantiles:
	:param get_bins: return heatmap bins
	:return: heatmap_df dataframe
	"""
	cut_1, var_1_bins = pandas.qcut(df[var_x], q=var_x_n_quantiles, retbins=True)
	cut_2, var_2_bins = pandas.qcut(df[var_y], q=var_y_n_quantiles, retbins=True)

	heatmap_df = pandas.pivot_table(index=[cut_1, cut_2], aggfunc={var_z: z_aggfunc}).unstack()
	seaborn.heatmap(heatmap_df, **kwargs)

	if get_bins:
		return heatmap_df, var_1_bins, var_2_bins
	else:
		return heatmap_df
